#!/usr/bin/env python
from __future__ import absolute_import

import boto3
import click
from click import ClickException
import json
import os
import platform
import sys
from base64 import b64decode
from botocore.exceptions import ClientError
from botocore.config import Config
from lambkin.aws import get_role_arn, get_event_rule_arn
from lambkin.aws import get_function_arn, get_region
from lambkin.runtime import get_sane_runtime, get_file_extension_for_runtime
from lambkin.runtime import get_language_name_for_runtime
from lambkin.template import render_template
from lambkin.ux import say
from lambkin.version import VERSION
from lambkin.virtualenv import create_virtualenv, run_in_virtualenv
from lambkin.zip import create_zip
import lambkin.metadata as metadata
from subprocess import check_output, CalledProcessError, STDOUT

lmbda = boto3.client('lambda', region_name=get_region())


@click.command(help='Make a new Lambda function from a basic template.')
@click.argument('function')
@click.option('--runtime', help='The language runtime to use. eg. "python2.7".')
def create(function, runtime):
    runtime = get_sane_runtime(runtime)

    ext = get_file_extension_for_runtime(runtime)
    func_dir = function
    func_file = os.path.join(func_dir, '%s.%s' % (function, ext))

    for path in (func_dir, func_file):
        if os.path.exists(path):
            raise ClickException('Path "%s" already exists.' % path)

    os.mkdir(func_dir)

    template_name = get_language_name_for_runtime(runtime)
    render_template(template_name, function,
                    output_filename="%s.%s" % (function, ext))
    if get_language_name_for_runtime(runtime) == 'python':
        create_virtualenv(function)
        render_template('requirements', function, output_filename='requirements.txt')
        render_template('gitignore-python', function, output_filename='.gitignore')
    else:
        render_template('makefile', function, output_filename='Makefile')
        render_template('gitignore', function, output_filename='.gitignore')

    our_metadata = {
        'function': function,
        'runtime': runtime,
        'language': get_language_name_for_runtime(runtime),
        'timeout': metadata.get('timeout')
    }
    metadata.write(subdirectory=function, **our_metadata)

    say('%s created as %s' % (function, func_file))


def get_published_function_names():
    """Return the names of all published functions."""
    return [f['FunctionName'] for f in lmbda.list_functions()['Functions']]


@click.command(name='list-published',
               help='List published Lambda functions.')
def list_published():
    for f in lmbda.list_functions()['Functions']:
        print f['FunctionName']


@click.command(help="Run the build process for a function.")
def build():
    runtime = metadata.get('runtime')
    language = get_language_name_for_runtime(runtime)
    if language == 'python':
        # Use virtualenv and pip
        if not os.path.isfile('venv'):
            create_virtualenv('.')
        print run_in_virtualenv('pip install -r requirements.txt')
    else:
        # Fall back to a Makefile.
        try:
            make_log = check_output(['make'], stderr=STDOUT)
            for line in make_log.rstrip().split("\n"):
                say(line)
        except CalledProcessError as e:
            for line in e.output.rstrip().split("\n"):
                say(line)
            raise ClickException('make failure')


@click.command(help='Publish a function to Lambda.')
@click.option('--description', help="Descriptive text in AWS Lamda.")
@click.option('--timeout', type=click.IntRange(min=1, max=300),
              help="Maximum time the function can run, in seconds.")
@click.option('--memory', type=click.IntRange(min=128, max=1536),
              help="Memory allocated to the function, in MiB.")
@click.option('--role', help="Lambda execution role. Default: lambda_basic_execution")
@click.option('--zip-file-path', help="Name of zip file that lambkin creates. Default: /tmp/lambkin-publish-<function>.zip.")
@click.option('--zip-file-only', is_flag=True, help="Produce zip file and exit without publishing.")
def publish(description, timeout, memory, role, zip_file_only, zip_file_path):
    runtime = metadata.get('runtime')
    function = metadata.get('function')

    if description:
        metadata.update(description=description)
    else:
        try:
            description = metadata.get('description')
        except KeyError:
            raise ClickException('Please provide a description with "--description"')

    if timeout:
        metadata.update(timeout=timeout)
    else:
        timeout = metadata.get('timeout')

    if memory:
        metadata.update(memory=memory)
    else:
        memory = metadata.get('memory')

    if role:
        metadata.update(role=role)
    else:
        role = metadata.get('role')

    zip_data = open(create_zip(zip_file_path)).read()

    if zip_file_only:
        return

    if function in get_published_function_names():
        # Push the latest code to the existing function in Lambda.
        update_code_response = lmbda.update_function_code(
            FunctionName=function,
            ZipFile=zip_data,
            Publish=True)

        # Update any settings for the function too.
        final_response = lmbda.update_function_configuration(
            FunctionName=function,
            Description=description,
            Role=get_role_arn(role),
            Timeout=timeout,
            MemorySize=memory)
        say('%s updated in Lambda' % function)
    else:  # we need to explictly create the function in AWS.
        final_response = lmbda.create_function(
            FunctionName=function,
            Description=description,
            Runtime=runtime,
            Role=get_role_arn(role),
            Handler='%s.handler' % function,
            Code={'ZipFile': zip_data},
            Timeout=timeout,
            MemorySize=memory)
        say('%s created in Lambda' % function)
    print json.dumps(final_response, sort_keys=True, indent=2)


@click.command(help='Run a published function.')
@click.option('--function', help="Defaults to the function in the current dir.")
def run(function):
    lmbda = boto3.client('lambda', region_name=get_region(),
                         config=Config(retries={'max_attempts': 1}, read_timeout=310))
    if not function:
        function = metadata.get('function')
    result = lmbda.invoke(FunctionName=function, LogType='Tail')
    log = b64decode(result['LogResult']).rstrip().split("\n")
    for line in log:
        say(line)
    print result['Payload'].read()


@click.command(help='Remove a function from Lambda.')
@click.option('--function', help="Defaults to the function in the current dir.")
def unpublish(function):
    if not function:
        function = metadata.get('function')
    lmbda.delete_function(FunctionName=function)
    say('%s unpublished' % (function))


@click.command(help='Schedule a function to run regularly.')
@click.option('--function', help="Defaults to the function in the current dir.")
@click.option('--rate', help='Execution rate. Like "6 minutes", or "1 day".')
@click.option('--cron', help='Cron schedule. Like "0 8 1 * ? *".')
def schedule(function, rate, cron):
    if not function:
        function = metadata.get('function')
    events = boto3.client('events', region_name=get_region())

    if (rate and cron):
        raise ClickException(
            'Please use either "--rate" or "--cron", not both.')
    if rate:
        schedule_expression = 'rate(%s)' % rate
    elif cron:
        schedule_expression = 'cron(%s)' % cron
    else:
        raise ClickException('Please provide "--rate" or "--cron".')

    try:
        lmbda.add_permission(
            FunctionName=function,
            StatementId='lambkin-allow-cloudwatch-invoke-%s' % function,
            Action='lambda:InvokeFunction',
            Principal='events.amazonaws.com',
            SourceArn=get_event_rule_arn('lambkin-cron-%s' % function),
        )
    except ClientError as e:
        if e.response['Error']['Code'] == 'ResourceConflictException':
            # The rule is already there. Carry on!
            pass
        else:
            raise e

    events.put_rule(
        Name='lambkin-cron-%s' % function,
        ScheduleExpression=schedule_expression,
        State='ENABLED',
        Description='Lambkin cron for %s Lambda function' % function,
    )

    response = events.put_targets(
        Rule='lambkin-cron-%s' % function,
        Targets=[{
            'Id': function,
            'Arn': get_function_arn(function),
        }]
    )

    print json.dumps(response, sort_keys=True, indent=2)


def main():
    if platform.system() == 'Windows':
        print "Lambkin doesn't run on Windows yet. Sorry."
        sys.exit(1)

    @click.group(invoke_without_command=True, no_args_is_help=True)
    @click.pass_context
    @click.option('--version', help='Show the version.', is_flag=True)
    def cli(ctx, version):
        if ctx.invoked_subcommand is None and version:
            click.echo(VERSION)

    subcommands = [create, list_published, build, publish, run, schedule,
                   unpublish]
    for cmd in subcommands:
        cli.add_command(cmd)
    cli()


if __name__ == '__main__':
    main()



import json
import os

metadata_file = 'metadata.json'
defaults = {
    'timeout': 60,
    'memory': 128,
    'role': 'lambda_basic_execution'
}


def get(key):
    """Return a single named property from the metadata.

    Defaults are returned for some properties if they are not stored in
    the metadata.json file.
    """
    try:
        return read()[key]
    except KeyError:
        return defaults[key]


def read():
    """Read metadata for a function from disk (metadata.json)."""
    if os.path.exists(metadata_file):
        with open(metadata_file) as f:
            return json.load(f)
    else:
        return {}


def write(subdirectory=None, **metadata):
    """Write keyword arguments, in JSON format, to metadata.json."""
    if subdirectory:
        path = os.path.join(subdirectory, metadata_file)
    else:
        path = metadata_file
    with open(path, 'w') as f:
        serialized = json.dumps(metadata, indent=4, sort_keys=True)
        f.write(serialized)


def update(**metadata):
    """Update the metadata file with any given values."""
    updated_metadata = read()
    for k, v in metadata.iteritems():
        updated_metadata[k] = metadata[k]
    write(**updated_metadata)



from __future__ import absolute_import

import os
import zipfile
import lambkin.metadata as metadata

# REF: http://docs.aws.amazon.com/lambda/latest/dg/lambda-python-how-to-create-deployment-package.html


def create_zip(zip_file_path):
    if not zip_file_path:
        function = metadata.get('function')
        zip_file_path = '/tmp/lambkin-publish-%s.zip' % function

    zip_file = zipfile.ZipFile(zip_file_path, 'w', zipfile.ZIP_DEFLATED)

    for root, dirs, files in os.walk('.'):
        site_dir = os.path.join('.', 'venv', 'lib', 'python2.7', 'site-packages')
        dist_dir = os.path.join('.', 'venv', 'lib', 'python2.7', 'dist-packages')
        for f in files:
            path = os.path.join(root, f)
            if path.endswith('.pyc'):
                pass
            elif root.startswith(site_dir):
                # Then strip the library dir, and put the file in the zip.
                trimmed_path = path[len(site_dir):]
                zip_file.write(path, trimmed_path)
            elif root.startswith(dist_dir):
                # Then strip the library dir, and put the file in the zip.
                trimmed_path = path[len(dist_dir):]
                zip_file.write(path, trimmed_path)
            elif root.startswith('./venv') or root.startswith('./.git'):
                # Then it's other junk that we don't want.
                pass
            else:
                # Not sure what this is. The function author probably put
                # it here for a reason, so make sure it goes in the zip.
                zip_file.write(path, path)
    zip_file.close()
    return zip_file_path



import os
from click import ClickException
from subprocess import call, check_output
from os.path import join


def have_virtualenv():
    try:
        with open(os.devnull, 'w') as null:
            status = call(['virtualenv', '--help'], stdout=null, stderr=null)
        return status is 0
    except OSError:
        return False


def create_virtualenv(function_name):
    if not have_virtualenv():
        raise ClickException('Lambkin needs virtualenv. Please install it.')
    check_output([
        'virtualenv', '--python=python2.7', join(function_name, 'venv')
    ])


def run_in_virtualenv(command):
    """Run a shell command with the current function's virtualenv active."""
    return check_output('. venv/bin/activate && %s' % command, shell=True)



import pystache
from os.path import join
from textwrap import dedent


def render_template(template_name, function_name, output_filename=None):
    if not output_filename:
        output_filename = template_name
    output = join(function_name, output_filename)
    expansions = {
        'function_name': function_name
    }
    with open(output, 'w') as dst:
        dst.write(pystache.render(templates[template_name], expansions))

templates = {
    'python': dedent(
        """
        # {{function_name}}.py: An AWS Lambda function.
        #
        # To run this function in Lambda, you need to bundle all the libraries you use.
        # Even something as mundane as:
        #   import requests
        # won't work by default.
        #
        # To install a library (like "requests") into your Lambda function, edit
        # the requirements.txt in your function's directory, then run:
        #
        #  lambkin build
        #
        # The files installed by pip are all "gitignored" by default, so they don't
        # get commited. In fact, almost everything is ignored, so if you want to add
        # and commit more than just your script, you'll need to explicitly list things
        # in "functions/{{function_name}}/.gitignore".
        #
        #
        #
        # Here is the entry point for Lambda. Execution starts here when an
        # event triggers us to run.
        #
        # Lambda passes us two objects, the event that triggered us, and a 'context'
        # object for this execution.

        def handler(event, context):
            # To log to Cloudwatch, just:
            print "anything you like."

            optional_return_value = {
                "hello": "World",
                "from": "Python"
            }

            return optional_return_value  # which often becomes a JSON response.
        """[1:]),

    'makefile': dedent("""
        {{function_name}}:
        	@echo "make" called for {{function_name}}
        	# npm install --prefix=. aws-sdk
        """[1:]),

    'nodejs': dedent(
        """
        // {{function_name}}.js: An AWS Lambda function.

        exports.handler = function(event, context, callback) {
          console.log("Running in Lambda...");

          // An abitrary object.
          result = {
            "hello": "World",
            "from": "Node.js"
          };

          // This example uses an NPM package. Look in the "Makefile" to see how to
          // install packages for your Lambda function.
          //
          // var yaml = require('js-yaml');
          // console.log(yaml.safeDump(result));

          // Lambda will emit the "result" object as the output of our function.
          callback(null, result);
        }
        """[1:]),

    'gitignore': dedent(
        """
        # If this Lambda function uses any libraries, they will be here in the
        # source directory.
        #
        # We don't want to commit copies of 3rd party libraries to our repository,
        # so by default we will ignore _everything_ in the source dir:
        *
        # ...except for some things we want to commit:
        !{{function_name}}.*
        !metadata.json
        !Makefile
        !.gitignore
        """[1:]),

    'gitignore-python': dedent(
        """
        /venv
        """[1:]),

    'requirements': dedent(
        """
        # Declare requirements for your Python Lambda function here.
        #
        # eg.
        # requests>=2
        """[1:])
}



from click import ClickException


def get_file_extension_for_runtime(runtime):
    """Return the correct file extension for a language runtime."""
    runtime = get_sane_runtime(runtime)
    if runtime.startswith('python'):
        return 'py'
    elif runtime.startswith('nodejs'):
        return 'js'


def get_language_name_for_runtime(runtime):
    if runtime.startswith('python'):
        return 'python'
    elif runtime.startswith('node'):
        return 'nodejs'


def get_sane_runtime(runtime):
    """Make a runtime string that Lambda will accept."""

    # Default to Python
    if runtime is None:
        runtime = 'python'

    # Ensure we specify a version, not just a language.
    if runtime == 'python':
        runtime = 'python2.7'
    if runtime == 'node' or runtime == 'nodejs':
        runtime = 'nodejs4.3'

    valid_runtimes = ['python2.7', 'nodejs4.3']
    if runtime in valid_runtimes:
        return runtime
    else:
        raise ClickException('Runtime "%s" is not supported' % runtime)



from click import echo


def say(message):
    """Write an info message to STDERR.

    We deliberatly send the message to STDERR to keep STDOUT clear for any
    JSON output we may return. Interleaving info messages would break the
    ability to pipe our output to a JSON parser.
    """
    echo(message, err=True)



from __future__ import absolute_import

from setuptools import setup
from lambkin.version import VERSION

import os.path


if(os.path.isfile('README.md')):
    import pypandoc
    with open('README.rst', 'w') as rst:
        rst.write(pypandoc.convert('README.md', 'rst', format='md'))

setup(
    name='lambkin',
    packages=['lambkin'],
    version=VERSION,
    description='CLI tool for managing functions in AWS Lambda.',
    long_description=open('README.rst').read(),
    author='Toby McLaughlin',
    author_email='toby@jarpy.net',
    url='https://github.com/jarpy/lambkin',
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX',
        'Development Status :: 4 - Beta',
        'Environment :: Console',
    ],
    entry_points={
        'console_scripts': [
            'lambkin=lambkin.lambkin:main',
        ],
    },
    install_requires=[
        'boto3',
        'click>=6,<7',
        'pystache'
    ]

)



import boto3


def get_region():
    """Return the configured AWS region, or "us-east-1" if undefined."""
    return boto3.session.Session().region_name or 'us-east-1'


def get_account_id():
    """Get the AWS account id for our current credentials."""
    return boto3.resource('iam').CurrentUser().arn.split(':')[4]


def get_iam_arn_prefix():
    return "arn:aws:iam::%s" % get_account_id()


def get_event_arn_prefix():
    return "arn:aws:events:%s:%s" % (get_region(), get_account_id())


def get_role_arn(role):
    """Given a short role name, return the ARN of the role.

    eg.

    get_role_arn('lambda_basic_execution')
    "arn:aws:iam::329487123:role/lambda-basic-execution"
    """
    role = boto3.resource('iam').Role(role)
    return role.arn


def get_event_rule_arn(rule):
    """Given a short event rule name, return the ARN of the rule.

    eg.

    get_event_rule_arn('allow-from-s3')
    "arn:aws:events:us-west-2:329487123:rule/allow-from-s3"
    """
    return "%s:rule/%s" % (get_event_arn_prefix(), rule)


def get_function_arn(function):
    """Given a short function name, return the ARN of function."""
    return 'arn:aws:lambda:%s:%s:function:%s' % (
        get_region(), get_account_id(), function
    )
